@@ -225,7 +225,7 @@ struct TTASFutexMutex {
 				return;
 			}
 
-			ut_delay(ut_rnd_interval(0, max_delay));
+			ut_delay(max_delay);
 		}
 
 		for (n_waits= 0;; n_waits++) {
@@ -362,7 +362,7 @@ struct TTASMutex {
 		uint32_t n_spins = 0;
 
 		while (!try_lock()) {
-			ut_delay(ut_rnd_interval(0, max_delay));
+			ut_delay(max_delay);
 			if (++n_spins == max_spins) {
 				os_thread_yield();
 				max_spins+= step;
@@ -516,7 +516,7 @@ struct TTASEventMutex {
 					sync_array_wait_event(sync_arr, cell);
 				}
 			} else {
-				ut_delay(ut_rnd_interval(0, max_delay));
+				ut_delay(max_delay);
 			}
 		}
 
